@@ -21,10 +21,4 @@
 
 """Python implementation of the Git file formats and protocols."""
 
-
-import client
-import protocol
-import repo
-import server
-
-__version__ = (0, 6, 1)
+__version__ = (0, 9, 5)
